from scrapy_ajax_utils.utils import add_settings_to_spider


def selenium_support(spider_cls):
    """Spider class decorator"""
    settings = {
        'DOWNLOADER_MIDDLEWARES': {'scrapy_ajax_utils.selenium.SeleniumDownloaderMiddleware': 600}
    }
    return add_settings_to_spider(spider_cls, settings)


# top import.
from .browser import Browser
from .request import SeleniumRequest
from .middleware import SeleniumDownloaderMiddleware



import logging
import threading

from scrapy import signals
from scrapy.http import Request, Response

from selenium.webdriver.support.ui import WebDriverWait
from selenium.common.exceptions import TimeoutException

from twisted.internet import threads, reactor
from twisted.python.threadpool import ThreadPool

from scrapy_ajax_utils.selenium.browser import Browser
from scrapy_ajax_utils.selenium.request import SeleniumRequest

logger = logging.getLogger(__name__)


class SeleniumDownloaderMiddleware(object):

    @classmethod
    def from_crawler(cls, crawler):
        settings = crawler.settings
        min_drivers = settings.get('SELENIUM_MIN_DRIVERS', 3)
        max_drivers = settings.get('SELENIUM_MAX_DRIVERS', 5)
        browser = _make_browser_from_settings(settings)
        dm = cls(browser, min_drivers, max_drivers)
        crawler.signals.connect(dm.spider_closed, signal=signals.spider_closed)
        return dm

    def __init__(self, browser, min_drivers, max_drivers):
        self._browser = browser
        self._drivers = set()
        self._data = threading.local()
        self._threadpool = ThreadPool(min_drivers, max_drivers)

    def process_request(self, request, spider):
        if not isinstance(request, SeleniumRequest):
            return

        if not self._threadpool.started:
            self._threadpool.start()

        return threads.deferToThreadPool(
            reactor, self._threadpool, self.download_by_driver, request, spider
        )

    def download_by_driver(self, request, spider):
        driver = self.get_driver()

        try:
            driver.get(request.url)
        except TimeoutException:
            spider.logger.info(f'page download timeout (with selenium): {request.url}')

        # XXX: Check cookies.
        # if request.cookies:
        #     if isinstance(request.cookies, list):
        #         for cookie in request.cookies:
        #             driver.add_cookie(cookie)
        #     else:
        #         for k, v in request.cookies.items():
        #             driver.add_cookie({'name': k, 'value': v})
        #     driver.get(request.url)

        if request.wait_until:
            WebDriverWait(driver, request.wait_time).until(request.wait_until)

        # Execute javascript code and put the result to meta.
        if request.script:
            request.meta['js_result'] = driver.execute_script(request.script)

        if request.handler:
            result = request.handler(driver, request, spider)
            if isinstance(result, (Request, Response)):
                return result

        return driver.current_response(request)

    def get_driver(self):
        try:
            driver = self._data.driver
        except AttributeError:
            driver = self._browser.driver()
            self._drivers.add(driver)
            self._data.driver = driver
        return driver

    def spider_closed(self):
        for driver in self._drivers:
            driver.quit()
        logger.debug('all webdriver closed.')
        self._threadpool.stop()


def _make_browser_from_settings(settings):
    headless = settings.getbool('SELENIUM_HEADLESS', True)
    disable_image = settings.get('SELENIUM_DISABLE_IMAGE', True)
    driver_name = settings.get('SELENIUM_DRIVER_NAME', 'chrome')
    executable_path = settings.get('SELENIUM_DRIVER_PATH')
    user_agent = settings.get('USER_AGENT')
    page_load_time_out = settings.get('SELENIUM_DRIVER_PAGE_LOAD_TIMEOUT', 30)
    return Browser(headless=headless,
                   disable_image=disable_image,
                   driver_name=driver_name,
                   executable_path=executable_path,
                   user_agent=user_agent,
                   page_load_time_out=page_load_time_out)



import copy


def add_settings_to_spider(spider_cls, settings: dict):
    s = copy.deepcopy(settings)
    cus_settings = spider_cls.custom_settings
    if cus_settings:
        for k, v in cus_settings.items():
            if isinstance(v, dict) and k in s:
                s[k].update(v)
            else:
                s[k] = v
    spider_cls.custom_settings = s
    return spider_cls



import scrapy


class SeleniumRequest(scrapy.Request):

    def __init__(self,
                 url,
                 callback=None,
                 wait_until=None,
                 wait_time=10,
                 script=None,
                 handler=None,
                 **kwargs):
        """
        :param url: target url to request.
        :param callback: ..
        :param wait_until: see usage in: `WebDriverWait.until`
        :param wait_time:
        :param script: javascript to execute
                    and the result will be save to response.meta
                    which key is 'js_result'

        :param handler: function to handle the driver
                    the function only accept 3 arguments:
                        driver: the instance of `scrapy_ajax_utils.selenium.browser._WebDriver`
                        request: current request
                        spider: current spider

                    and it will execute when:
                        url is got
                        and wait until passed (if exist)
                        and the javascript be executed (if exist)

                    if the function has return value and the value is instance of
                    `scrapy.Request` or `scrapy.Response`, it will be returned in the
                    'process_request' method of downloader middleware immediately

        :param kwargs: keyword arguments to pass to `scrapy.Request`
        """
        self.wait_until = wait_until
        self.wait_time = wait_time
        self.script = script
        self.handler = handler
        super().__init__(url, callback, **kwargs)



from scrapy.http import HtmlResponse
from scrapy.downloadermiddlewares.cookies import CookiesMiddleware

from selenium import webdriver
from selenium.webdriver.remote.webdriver import WebDriver as RemoteWebDriver

_format_cookie = CookiesMiddleware()._format_cookie


class Browser(object):
    """Browser to make drivers"""

    support_driver_map = {
        'firefox': webdriver.Firefox,
        'chrome': webdriver.Chrome
    }

    def __init__(self, driver_name='chrome', executable_path=None, options=None, page_load_time_out=30, **opt_kw):
        assert driver_name in self.support_driver_map, f'{driver_name} not be supported!'
        self.driver_name = driver_name
        self.executable_path = executable_path
        self.page_load_time_out = page_load_time_out
        if options is not None:
            self.options = options
        else:
            self.options = make_options(self.driver_name, **opt_kw)

    def driver(self):
        kwargs = {'executable_path': self.executable_path, 'options': self.options}
        # Close log file, only works for windows.
        if self.driver_name == 'firefox':
            kwargs['service_log_path'] = 'nul'
        driver = self.support_driver_map[self.driver_name](**kwargs)
        driver.set_page_load_timeout(self.page_load_time_out)
        self.prepare_driver(driver)
        return _WebDriver(driver)

    def prepare_driver(self, driver):
        if isinstance(driver, webdriver.Chrome):
            # Remove `window.navigator.webdriver`.
            driver.execute_cdp_cmd("Page.addScriptToEvaluateOnNewDocument", {
                "source": """
                Object.defineProperty(navigator, 'webdriver', {
                  get: () => undefined
                })
              """
            })


class _WebDriver(object):
    """As agent of the webdriver.WebDriver"""
    def __init__(self, driver: RemoteWebDriver):
        self._driver = driver

    def __getattr__(self, item):
        return getattr(self._driver, item)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self._driver.quit()

    def current_response(self, request):
        cookies = filter(None, (_format_cookie(c, request) for c in self.get_cookies()))
        headers = {'Set-Cookie': cookies}
        return HtmlResponse(self.current_url,
                            body=str.encode(self.page_source),
                            encoding='utf-8',
                            request=request,
                            headers=headers)


def make_options(driver_name, headless=True, disable_image=True, user_agent=None):
    if driver_name == 'chrome':
        options = webdriver.ChromeOptions()
        options.headless = headless
        options.add_argument('--disable-gpu')
        if user_agent:
            options.add_argument(f"--user-agent={user_agent}")
        if disable_image:
            options.add_experimental_option('prefs', {'profile.default_content_setting_values': {'images': 2}})
        options.add_experimental_option('excludeSwitches', ['enable-automation', ])
        return options

    elif driver_name == 'firefox':
        options = webdriver.FirefoxOptions()
        options.headless = headless
        if disable_image:
            options.set_preference('permissions.default.image', 2)
        if user_agent:
            options.set_preference('general.useragent.override', user_agent)
        return options
